# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from unittest import mock

import pytest

from croud.api import Client, RequestMethod
from croud.users.commands import transform_roles_list
from tests.util import assert_rest, call_command

pytestmark = pytest.mark.usefixtures("config")


@mock.patch.object(Client, "request", return_value=({}, None))
def test_users_list(mock_request):
    call_command("croud", "users", "list")
    assert_rest(mock_request, RequestMethod.GET, "/api/v2/users/", params=None)


@mock.patch.object(Client, "request", return_value=({}, None))
def test_users_list_no_org(mock_request, capsys):
    call_command("croud", "users", "list", "--no-org")
    assert_rest(
        mock_request, RequestMethod.GET, "/api/v2/users/", params={"no-roles": "1"}
    )
    _, err = capsys.readouterr()
    assert "The --no-org argument is deprecated. Please use --no-roles instead." in err


@mock.patch.object(Client, "request", return_value=({}, None))
def test_users_list_no_roles(mock_request):
    call_command("croud", "users", "list", "--no-roles")
    assert_rest(
        mock_request, RequestMethod.GET, "/api/v2/users/", params={"no-roles": "1"}
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_users_list_no_org_no_roles(mock_request, capsys):
    call_command("croud", "users", "list", "--no-roles", "--no-org")
    assert_rest(
        mock_request, RequestMethod.GET, "/api/v2/users/", params={"no-roles": "1"}
    )
    _, err = capsys.readouterr()
    assert "The --no-org argument is deprecated. Please use --no-roles instead." in err


@mock.patch.object(Client, "request", return_value=({}, None))
def test_users_roles_list(mock_request):
    call_command("croud", "users", "roles", "list")
    assert_rest(mock_request, RequestMethod.GET, "/api/v2/roles/")


@mock.patch.object(Client, "request", return_value=({}, None))
def test_transform_roles_list(mock_request):
    user = {
        "organization_roles": [
            {"organization_id": "org-1", "role_fqn": "org_admin"},
            {"organization_id": "org-2", "role_fqn": "org_member"},
            {"organization_id": "org-3", "role_fqn": "org_member"},
        ],
        "project_roles": [
            {"project_id": "project-1", "role_fqn": "project_admin"},
            {"project_id": "project-2", "role_fqn": "project_member"},
            {"project_id": "project-3", "role_fqn": "project_member"},
        ],
    }
    response = transform_roles_list("organization_id")(user["organization_roles"])
    assert response == "org-1: org_admin,\norg-2: org_member,\norg-3: org_member"
    response = transform_roles_list("project_id")(user["project_roles"])
    assert response == (
        "project-1: project_admin,\n"
        "project-2: project_member,\n"
        "project-3: project_member"
    )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from unittest import mock

import pytest
import yaml

import croud.config
from croud.config.configuration import Configuration
from croud.config.util import clean_dict
from tests.util import call_command


def test_config_add_profile(config, capsys):
    call_command(
        "croud",
        "config",
        "profiles",
        "add",
        "new-profile",
        "--endpoint",
        "http://localhost:8000",
        "--region",
        "new-region",
    )
    assert "format" not in config.profiles["new-profile"]
    assert config.profiles["new-profile"]["region"] == "new-region"

    _, err = capsys.readouterr()
    assert "Added profile 'new-profile'" in err

    call_command(
        "croud",
        "config",
        "profiles",
        "add",
        "newer-profile",
        "--endpoint",
        "http://localhost:8000",
        "--format",
        "json",
    )
    assert config.profiles["newer-profile"]["format"] == "json"
    assert "region" not in config.profiles["newer-profile"]

    _, err = capsys.readouterr()
    assert "Added profile 'newer-profile'" in err


def test_config_add_duplicate_profile(config, capsys):
    with pytest.raises(SystemExit) as exc_info:
        call_command(
            "croud",
            "config",
            "profiles",
            "add",
            "cratedb.cloud",
            "--endpoint",
            "http://localhost:8000",
        )
    assert exc_info.value.code == 1
    _, err = capsys.readouterr()
    assert "Failed to add profile 'cratedb.cloud'" in err


def test_config_current_profile(config, capsys):
    call_command("croud", "config", "profiles", "current")
    out, _ = capsys.readouterr()
    assert config.name in out


def test_config_remove_profile(config, capsys):
    call_command("croud", "config", "profiles", "remove", "cratedb.cloud")
    assert "cratedb.cloud" not in config.profiles

    _, err = capsys.readouterr()
    assert "Removed profile 'cratedb.cloud'" in err


def test_config_remove_current_profile(config, capsys):
    with pytest.raises(SystemExit) as exc_info:
        call_command("croud", "config", "profiles", "remove", config.name)
    assert exc_info.value.code == 1
    assert config.name in config.profiles

    _, err = capsys.readouterr()
    assert f"Failed to remove profile '{config.name}'" in err


def test_config_set_profile(config, capsys):
    call_command(
        "croud",
        "config",
        "profiles",
        "add",
        "new-profile",
        "--endpoint",
        "http://localhost:8000",
    )
    assert config.name != "new-profile"
    call_command("croud", "config", "profiles", "use", "new-profile")
    assert config.name == "new-profile"

    _, err = capsys.readouterr()
    assert "Switched to profile 'new-profile'" in err


def test_config_set_profile_does_not_exist(config, capsys):
    with pytest.raises(SystemExit) as exc_info:
        call_command("croud", "config", "profiles", "use", "invalid")
    assert exc_info.value.code == 1

    _, err = capsys.readouterr()
    assert "Failed to switch to profile 'invalid'" in err
    assert "Make sure the profile exists." in err


def test_config_show(config, capsys):
    call_command("croud", "config", "show")
    out, err = capsys.readouterr()

    assert out == yaml.safe_dump(clean_dict(config.config)) + "\n"
    assert "Configuration file" in err


def test_invalid_config(capsys, tmp_path):
    with open(tmp_path / "invalid.yaml", "w") as fp:
        # Write a legacy format config
        fp.write(
            """\
auth:
  contexts:
    dev:
      organization_id: ''
      token: ''
    local:
      organization_id: ''
      token: ''
    prod:
      organization_id: ''
      token: ''
  current_context: prod
output_fmt: table
region: bregenz.a1
"""
        )

    invalid_config = Configuration("invalid.yaml", tmp_path)

    with mock.patch.object(croud.config, "_CONFIG", invalid_config):
        with pytest.raises(SystemExit) as exc_info:
            call_command("croud", "me")
    assert exc_info.value.code == 1

    _, err = capsys.readouterr()
    assert (
        "Your configuration file is incompatible with the current version of croud."
        in err
    )
    assert (
        f"Please delete the file '{invalid_config._file_path}' or update it manually."
        in err
    )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace
from typing import Optional

from croud.api import Client
from croud.config import CONFIG
from croud.printer import print_error, print_info, print_warning
from croud.server import Server
from croud.util import can_launch_browser, open_page_in_browser


def login_path(idp: str = None) -> str:
    extra_part = f"{idp}/" if idp else ""
    return f"/oauth2/{extra_part}login?cli=true"


def get_org_id() -> Optional[str]:
    client = Client(CONFIG.endpoint, token=CONFIG.token, region=CONFIG.region)
    data, error = client.get("/api/v2/users/me/")
    if data and not error:
        return data.get("organization_id")
    return None


def login(args: Namespace) -> None:
    """
    Performs an OAuth2 Login to CrateDB Cloud
    """

    if not can_launch_browser():
        print_error("Login only works with a valid browser installed.")
        exit(1)

    server = Server(CONFIG.set_current_auth_token).start_in_background()

    open_page_in_browser(CONFIG.endpoint + login_path(args.idp))
    print_info("A browser tab has been launched for you to login.")
    try:
        # Wait for the user to login. They'll be redirected to the `SetTokenHandler`
        # which will set the token in the configuration.
        server.wait_for_shutdown()
    except (KeyboardInterrupt, SystemExit):
        print_warning("Login cancelled.")
    else:
        organization_id = get_org_id()
        CONFIG.set_organization_id(CONFIG.name, organization_id)
        print_info("Login successful.")
        print_info(f"Current profile: {CONFIG.name}")
        print_info(f"API endpoint   : {CONFIG.endpoint}")
        print_info(f"Organization ID: {organization_id}")



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import uuid
from unittest import mock

from croud.api import Client, RequestMethod
from tests.util import assert_rest, call_command, gen_uuid


@mock.patch.object(Client, "request", return_value=({}, None))
def test_subscriptions_list(mock_request):
    call_command("croud", "subscriptions", "list")
    assert_rest(mock_request, RequestMethod.GET, "/api/v2/subscriptions/")


@mock.patch.object(Client, "request", return_value=({}, None))
def test_subscriptions_list_with_organization_id(mock_request):
    org_id = gen_uuid()
    call_command("croud", "subscriptions", "list", "--org-id", org_id)
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/organizations/{org_id}/subscriptions/",
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_subscriptions_get(mock_request):
    id = str(uuid.uuid4())
    call_command("croud", "subscriptions", "get", id)
    assert_rest(mock_request, RequestMethod.GET, f"/api/v2/subscriptions/{id}/")


@mock.patch.object(Client, "request", return_value=({}, None))
def test_subscriptions_create(mock_request):
    call_command(
        "croud",
        "subscriptions",
        "create",
        "--type",
        "contract",
        "--org-id",
        "organization-id",
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/subscriptions/",
        body={"type": "contract", "organization_id": "organization-id"},
    )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace

from croud.api import Client
from croud.config import CONFIG, get_output_format
from croud.printer import print_error, print_response
from croud.util import org_id_config_fallback, require_confirmation


def organizations_create(args: Namespace) -> None:
    client = Client.from_args(args)
    if args.plan_type:
        body = {"name": args.name, "plan_type": args.plan_type}
    else:
        body = {"name": args.name}

    data, errors = client.post("/api/v2/organizations/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "plan_type"],
        success_message="Organization created.",
        output_fmt=get_output_format(args),
    )


@org_id_config_fallback
def organizations_edit(args: Namespace) -> None:
    client = Client.from_args(args)
    body = {}
    if args.plan_type:
        body["plan_type"] = args.plan_type
    if args.name:
        body["name"] = args.name
    if not body:
        print_error("No input arguments found.")
        exit(1)

    data, errors = client.put(f"/api/v2/organizations/{args.org_id}/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "plan_type"],
        success_message="Organization edited.",
        output_fmt=get_output_format(args),
    )


def organizations_get(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get(f"/api/v2/organizations/{args.id}/")
    print_response(
        data=data,
        errors=errors,
        output_fmt=get_output_format(args),
    )


def organizations_list(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get("/api/v2/organizations/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "plan_type"],
        output_fmt=get_output_format(args),
    )


@org_id_config_fallback
@require_confirmation(
    "Are you sure you want to delete the organization?",
    cancel_msg="Organization deletion cancelled.",
)
def organizations_delete(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(f"/api/v2/organizations/{args.org_id}/")
    print_response(
        data=data,
        errors=errors,
        success_message="Organization deleted.",
        output_fmt=get_output_format(args),
    )

    if errors is None and args.org_id == CONFIG.organization:
        CONFIG.set_organization_id(CONFIG.name, None)



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace

from croud.api import Client
from croud.config import get_output_format
from croud.printer import print_error, print_raw, print_response, print_success
from croud.util import require_confirmation


def regions_list(args: Namespace) -> None:
    """
    Prints the available regions
    """

    client = Client.from_args(args)
    if args.org_id:
        url = f"/api/v2/organizations/{args.org_id}/regions/"
    else:
        url = "/api/v2/regions/"
    data, errors = client.get(url)
    print_response(
        data=data,
        errors=errors,
        keys=["name", "description"],
        output_fmt=get_output_format(args),
    )


@require_confirmation(
    "Creating a region is an experimental feature. Are you sure you want to proceed?",  # noqa
    cancel_msg="Region creation cancelled.",
)
def regions_create(args: Namespace) -> None:
    """
    Creates a new region
    """
    client = Client.from_args(args)
    body = {"description": args.description, "organization_id": args.org_id}

    # Add optional parameters only when present
    if args.aws_bucket:
        body["aws_bucket"] = args.aws_bucket
    if args.aws_region:
        body["aws_region"] = args.aws_region

    region, region_errors = client.post("/api/v2/regions/", body=body)

    print_response(
        data=region,
        errors=region_errors,
        keys=["name", "description"],
        output_fmt=get_output_format(args),
    )

    if not region_errors and region and "name" in region:
        install_token, install_token_errors = client.get(
            f"/api/v2/regions/{region['name']}/install-token/"
        )

        if not install_token_errors and install_token and "token" in install_token:
            print_success("You have successfully created a region.")

            print_raw(
                [
                    "",
                    "To install the edge region run the following command:",
                    "",
                    f"  $ bash <( wget -qO- {client.base_url}/edge/cratedb-cloud-edge.sh) {install_token['token']}",  # noqa
                    "",
                ],
            )
        else:
            print_response(
                data=install_token,
                errors=install_token_errors,
                keys=["token"],
                output_fmt=get_output_format(args),
            )


@require_confirmation(
    "Deleting a region is an experimental feature. Are you sure you want to proceed?",  # noqa
    cancel_msg="Region deletion cancelled.",
)
def regions_delete(args: Namespace) -> None:
    """
    Deletes a new region
    """
    client = Client.from_args(args)
    region_name = args.name

    regions, errors = client.get("/api/v2/regions/")
    region = next((r for r in (regions or []) if r["name"] == region_name), None)

    if not region:
        print_error(f"The region {region_name} does not exist.")
        return

    # Check edge region status
    if region["is_edge_region"] and region["status"] == "UP":
        print_response(
            data=region,
            errors={
                "message": (
                    "Your region is still connected to CrateDB Cloud. Please uninstall "
                    "the CrateDB Edge stack from the region before deleting it by "
                    "running the script below:\nbash <(wget -qO- "
                    f"{client.base_url}/edge/uninstall-cratedb-cloud-edge.sh)"
                ),
                "errors": {},
            },
            output_fmt=get_output_format(args),
        )
        return

    data, region_errors = client.delete(f"/api/v2/regions/{region_name}/")

    if region_errors:
        print_response(
            data=data,
            errors={
                "message": region_errors.get("message"),
                "errors": {"related_resources": region_errors.get("related_resources")},
            },
            output_fmt=get_output_format(args),
        )
    else:
        print_response(
            data=data,
            errors=region_errors,
            success_message="You have successfully deleted a region.",
            output_fmt=get_output_format(args),
        )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace

from croud.api import Client
from croud.config import get_output_format
from croud.printer import print_response, print_warning


def transform_roles_list(key):
    def _transform(field):
        return ",\n".join(f"{r[key]}: {r['role_fqn']}" for r in field)

    return _transform


def users_list(args: Namespace) -> None:
    client = Client.from_args(args)
    if args.no_org:
        print_warning(
            "The --no-org argument is deprecated. Please use --no-roles instead."
        )

    no_roles = {"no-roles": "1"} if (args.no_roles or args.no_org) else None
    data, errors = client.get("/api/v2/users/", params=no_roles)
    print_response(
        data=data,
        errors=errors,
        output_fmt=get_output_format(args),
        keys=["uid", "email", "username", "organization_roles", "project_roles"],
        transforms={
            "organization_roles": transform_roles_list("organization_id"),
            "project_roles": transform_roles_list("project_id"),
        },
    )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import abc
import functools
import json
import sys
from typing import Any, Callable, Dict, List, Optional, Type, Union

import yaml
from colorama import Fore, Style
from tabulate import tabulate

from croud.tools.spinner import HALO
from croud.typing import JsonDict


def print_format(
    rows: Union[List[JsonDict], JsonDict],
    format: str,
    keys: Optional[List[str]] = None,
    transforms: Optional[Dict[str, Callable[[Any], Any]]] = None,
) -> None:
    try:
        Printer = PRINTERS[format]
    except KeyError:
        print_error("This print method is not supported.")
        exit(1)
    printer = Printer(keys, transforms)
    printer.print_rows(rows)


def print_response(
    data,
    errors,
    output_fmt,
    success_message: str = None,
    keys: List[str] = None,
    transforms: Dict[str, Callable[[Any], Any]] = None,
):
    if errors:
        if "message" in errors:
            print_error(errors["message"])
            if "errors" in errors:
                print_format(errors["errors"], "json")
        else:
            print_format(errors, "json")
        return

    if data is None:
        message = success_message or "Success."
        print_success(message)
        return

    print_format(data, output_fmt, keys, transforms)

    if data and success_message is not None:
        message = success_message
        print_success(message)


def print_raw(text: Union[List[str], str]):
    HALO.stop()
    if type(text) is list:
        text = "\n".join(text)
    print(text, file=sys.stderr)


def print_error(text: str):
    HALO.stop()
    print(Fore.RED + "==> Error: " + Style.RESET_ALL + text, file=sys.stderr)


def print_info(text: str):
    HALO.stop()
    print(Fore.CYAN + "==> Info: " + Style.RESET_ALL + text, file=sys.stderr)


def print_warning(text: str):
    HALO.stop()
    print(Fore.YELLOW + "==> Warning: " + Style.RESET_ALL + text, file=sys.stderr)


def print_success(text: str):
    HALO.stop()
    print(Fore.GREEN + "==> Success: " + Style.RESET_ALL + text, file=sys.stderr)


class FormatPrinter(abc.ABC):
    def __init__(
        self,
        keys: Optional[List[str]] = None,
        transforms: Optional[Dict[str, Callable[[Any], Any]]] = None,
    ):

        self.keys = keys
        self.transforms = transforms or {}

    def print_rows(self, rows: Union[List[JsonDict], JsonDict]) -> None:
        # Explicitly stop & clear the spinner
        HALO.stop()
        print(self.format_rows(rows))

    @abc.abstractmethod
    def format_rows(self, rows: Union[List[JsonDict], JsonDict]) -> str:
        raise NotImplementedError()


class JsonFormatPrinter(FormatPrinter):
    def format_rows(self, rows: Union[List[JsonDict], JsonDict]) -> str:
        return json.dumps(rows, sort_keys=False, indent=2)


class TableFormatPrinter(FormatPrinter):

    display_all_columns = False

    def format_rows(self, rows: Union[List[JsonDict], JsonDict]) -> str:
        if rows is None:
            return ""

        if not isinstance(rows, list):
            rows = [rows]

        if self.keys and (not self.display_all_columns or len(rows) == 0):
            filter_record = functools.partial(self._filter_record, keys=self.keys)
            rows = list(map(filter_record, rows))

        # ensure that keys are mapped to their values
        # e.g. Rows, generated with
        # [{"a": "foo", "b": 1}, {"b": 2, "a": "bar"}]
        # should get printed same as
        # [{"a": "foo", "b": 1}, {"a": "bar", "b": "1"}]
        # +-----+-----+
        # | a   |   b |
        # |-----+-----|
        # | foo |   1 |
        # |-----+-----|
        # | bar |   2 |
        # +-----+-----+

        headers = list(map(str, rows[0].keys())) if len(rows) else self.keys
        if headers is None:
            return ""

        values = [
            [
                self.transforms.get(header, TableFormatPrinter._identity_transform)(
                    row[header]
                )
                for header in headers
            ]
            for row in rows
        ]

        return tabulate(values, headers=headers, tablefmt="psql", missingval="NULL")

    def _filter_record(self, data: dict, keys: List[str]) -> JsonDict:
        # Ensure that we print keys in the order they are defined
        ret = {}
        for key in keys:
            if key in data:
                ret[key] = data[key]
        return ret

    @staticmethod
    def _identity_transform(field):
        """transform field for displaying"""
        if isinstance(field, (list, dict)):
            return json.dumps(field, sort_keys=True, ensure_ascii=False)
        elif isinstance(field, bool):
            return "TRUE" if field else "FALSE"
        else:
            return field


class WideTableFormatPrinter(TableFormatPrinter):
    """
    Prints data in tabular format without any field restrictions, displaying all the
    properties of the payload.
    """

    display_all_columns = True


class YamlFormatPrinter(FormatPrinter):
    def format_rows(self, rows: Union[List[JsonDict], JsonDict]) -> str:
        return yaml.dump(rows, default_flow_style=False)


PRINTERS: Dict[str, Type[FormatPrinter]] = {
    "json": JsonFormatPrinter,
    "table": TableFormatPrinter,
    "wide": WideTableFormatPrinter,
    "yaml": YamlFormatPrinter,
}



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.


class AsyncOperationNotFound(Exception):
    pass



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from pathlib import Path
from typing import Any, Dict, Optional

import yaml
from appdirs import user_config_dir
from marshmallow import ValidationError

from croud.config.exceptions import InvalidConfiguration, InvalidProfile
from croud.config.schemas import ConfigSchema, ProfileSchema
from croud.config.types import ConfigurationType, ProfileType

DEFAULT_CONFIGURATION = """\
default-format: table
current-profile: cratedb.cloud
profiles:
  cratedb.cloud:
    auth-token: NULL
    endpoint: https://console.cratedb.cloud
    region: _any_
"""


class Configuration:
    """
    The state representation class of the configuration file.

    There should only be a single instance of this class (representing the
    ``croud.yaml`` configuration file in the user's configuration directory,
    however for testing purposes, the name and path of the configuration file
    are configurable.

    When using the configuration, please use the :ref:`croud.config.CONFIG`
    instead.

    The configuration loads the contents of the configuration file lazily upon
    first access of the ``config`` property.

    The configuration is written to disk only whenever modifications are made
    to the in-memory state using the "public" methods.

    If there is no configuration file on disk, the state will be populated with
    the default configuration.
    """

    def __init__(self, name: str, path: Optional[Path] = None):
        self._config_dir = path or Path(user_config_dir("Crate"))
        self._file_path = self._config_dir / name
        self._config: Optional[ConfigurationType] = None
        self._schema = ConfigSchema()

    @property
    def config(self) -> ConfigurationType:
        if not self._config:
            self._config = self.load()
        return self._config

    @property
    def name(self) -> str:
        return self.config["current-profile"]  # type: ignore

    @property
    def endpoint(self) -> str:
        return self.profile["endpoint"]  # type: ignore

    @property
    def format(self) -> str:
        return self.profile.get("format", self.config["default-format"])  # type: ignore

    @property
    def token(self) -> Optional[str]:
        return self.profile["auth-token"]  # type: ignore

    @property
    def region(self) -> Optional[str]:
        return self.profile.get("region", None)  # type: ignore

    @property
    def organization(self) -> Optional[str]:
        return self.profile.get("organization-id")  # type: ignore

    @property
    def profile(self) -> ProfileType:
        return self.profiles[self.name]  # type: ignore

    @property
    def profiles(self) -> Dict[str, ProfileType]:
        return self.config["profiles"]  # type: ignore

    def load(self) -> ConfigurationType:
        if not self._file_path.exists():
            data = yaml.safe_load(DEFAULT_CONFIGURATION)
        else:
            with open(self._file_path, "r") as fp:
                data = yaml.safe_load(fp)
        # self._schema.load() will evaluate the correctness of the
        # configuration
        try:
            return self._schema.load(data)
        except ValidationError:
            raise InvalidConfiguration(
                f"{self._file_path} is not a valid configuration."
            ) from None
            exit(1)

    def is_valid(self):
        try:
            self.config
        except InvalidConfiguration:
            return False
        else:
            return True

    def dump(self) -> None:
        # make sure the config is in memory before we open the file for writing
        data = self.config
        self._config_dir.mkdir(mode=0o700, parents=True, exist_ok=True)
        with open(self._file_path, "w") as fp:
            yaml.safe_dump(data, fp)

    def _set_profile_option(self, profile: str, attr: str, value: Any) -> None:
        if profile not in self.profiles:
            raise InvalidProfile(profile)
        data = self.profiles[profile]
        data[attr] = value
        self.update_profile(profile, data)

    def set_organization_id(self, profile: str, value: str) -> None:
        self._set_profile_option(profile, "organization-id", value)

    def set_current_organization_id(self, value: str) -> None:
        self.set_organization_id(self.name, value)

    def set_auth_token(self, profile: str, value: str) -> None:
        self._set_profile_option(profile, "auth-token", value)

    def set_current_auth_token(self, value: str) -> None:
        self.set_auth_token(self.name, value)

    def set_format(self, profile: str, value: str) -> None:
        self._set_profile_option(profile, "format", value)

    def set_current_format(self, value: str) -> None:
        self.set_format(self.name, value)

    def add_profile(self, profile: str, *, endpoint: str, **kwargs) -> None:
        if profile in self.profiles:
            raise InvalidProfile(profile)
        data = {"auth-token": None, "endpoint": endpoint}  # required fields
        data.update(kwargs)  # optional fields
        self.update_profile(profile, data)

    def update_profile(self, profile: str, data: Dict) -> None:
        profile_schema = ProfileSchema()
        self.profiles[profile] = profile_schema.dump(data)
        self.dump()

    def remove_profile(self, profile: str) -> None:
        if profile not in self.profiles or profile == self.name:
            raise InvalidProfile(profile)
        del self.profiles[profile]
        self.dump()

    def use_profile(self, profile: str) -> None:
        if profile not in self.profiles:
            raise InvalidProfile(profile)
        self._config["current-profile"] = profile  # type: ignore
        self.dump()



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import enum
from argparse import Namespace
from platform import python_version
from typing import Any, Callable, Dict, Optional, Tuple

import requests
from yarl import URL

import croud
from croud.config import CONFIG
from croud.printer import print_error, print_info

ResponsePair = Tuple[Optional[Dict], Optional[Dict]]


class RequestMethod(enum.Enum):
    DELETE = "delete"
    GET = "get"
    PATCH = "patch"
    POST = "post"
    PUT = "put"


def noop(*args, **kwargs):
    pass


class Client:
    """
    A client for CrateDB Cloud API requests
    """

    def __init__(
        self,
        endpoint: str,
        *,
        token: str = None,
        on_token: Callable = None,
        region: str = None,
        sudo: bool = False,
        _verify_ssl: bool = True,
    ):
        """
        :param str endpoint:
          CrateDB Cloud API endpoint
        :param str token:
          The authentication token used as session cookie
        :param Callable on_token:
          A setter method that is called when a new auth token is retrieved
        :param str region:
          Region to get data from or to which to deploy (defines the
          ``X-Region`` HTTP header value)
        :param bool sudo:
          Whether or not to make requests as superuser (defines the
          ``X-Auth-Sudo`` HTTP header value)
        :param bool _verify_ssl:
          A private variable that must only be used during tests!
        """

        self.base_url = URL(endpoint)
        self._token = token
        self._on_token = on_token or noop

        self.session = requests.Session()
        self.session.cookies["session"] = self._token
        if _verify_ssl is False:
            self.session.verify = False
        if region:
            self.session.headers["X-Region"] = region
        if sudo:
            self.session.headers["X-Auth-Sudo"] = "1"

        ua = f"Croud/{croud.__version__} Python/{python_version()}"
        self.session.headers["User-Agent"] = ua

    @staticmethod
    def from_args(args: Namespace) -> "Client":
        return Client(
            CONFIG.endpoint,
            token=CONFIG.token,
            on_token=CONFIG.set_current_auth_token,
            region=args.region or CONFIG.region,
            sudo=args.sudo,
        )

    def request(
        self,
        method: RequestMethod,
        endpoint: str,
        *,
        params: dict = None,
        body: dict = None,
    ):
        # When logging out, the Gateway may respond with a redirect in case the
        # session is still valid and the IDP identifier is present.
        # We need to follow that redirect in order to fully log out the user!
        kwargs: Dict[str, Any] = {
            "allow_redirects": "logout" in endpoint,
        }
        if params is not None:
            kwargs["params"] = params
        if body is not None:
            kwargs["json"] = body

        try:
            response = self.session.request(
                method.value, str(self.base_url.with_path(endpoint)), **kwargs
            )
        except requests.RequestException as e:
            message = (
                f"Failed to perform request on '{e.request.url}'. "
                f"Original error was: '{e}'"
            )
            return None, {"message": message, "success": False}

        if response.is_redirect:
            redirect_url = URL(response.headers["Location"])
            print_error("Unauthorized.")
            if redirect_url.query.get("rd", "") == endpoint:  # login redirect
                # When use is unauthorized, the API request returns a 302
                # redirect to `/?rd=<api_endpoint>`
                print_info("Use `croud login` to login to CrateDB Cloud.")
            else:
                print_error("Oops. Something unexpected happened.")
            exit(1)

        # Refresh a previously provided token because it has timed out
        response_token = response.cookies.get("session")
        if response_token and response_token != self._token:
            self._token = response_token
            self._on_token(response_token)

        return self.decode_response(response)

    def delete(
        self, endpoint: str, *, params: dict = None, body: dict = None
    ) -> ResponsePair:
        return self.request(RequestMethod.DELETE, endpoint, params=params, body=body)

    def get(self, endpoint: str, *, params: dict = None) -> ResponsePair:
        return self.request(RequestMethod.GET, endpoint, params=params)

    def patch(
        self, endpoint: str, *, params: dict = None, body: dict = None
    ) -> ResponsePair:
        return self.request(RequestMethod.PATCH, endpoint, params=params, body=body)

    def post(
        self, endpoint: str, *, params: dict = None, body: dict = None
    ) -> ResponsePair:
        return self.request(RequestMethod.POST, endpoint, params=params, body=body)

    def put(
        self, endpoint: str, *, params: dict = None, body: dict = None
    ) -> ResponsePair:
        return self.request(RequestMethod.PUT, endpoint, params=params, body=body)

    def decode_response(self, resp: requests.Response) -> ResponsePair:
        if resp.status_code == 204:
            # response is empty
            return None, None

        try:
            # API always returns JSON, unless there's an unhandled server error
            body = resp.json()
        except ValueError:
            body = {"message": f"{resp.status_code} - {resp.reason}", "success": False}

        if resp.status_code >= 400:
            return None, body
        else:
            return body, None



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.
import time
from argparse import Namespace
from typing import Dict, Optional

import bitmath

from croud.api import Client
from croud.clusters.exceptions import AsyncOperationNotFound
from croud.config import get_output_format
from croud.printer import print_error, print_info, print_response, print_success
from croud.tools.spinner import HALO
from croud.util import require_confirmation


def clusters_get(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get(f"/api/v2/clusters/{args.id}/")
    print_response(
        data=data,
        errors=errors,
        output_fmt=get_output_format(args),
    )


def clusters_list(args: Namespace) -> None:
    if args.org_id:
        url = f"/api/v2/organizations/{args.org_id}/clusters/"
    else:
        url = "/api/v2/clusters/"

    params = {}
    if args.project_id:
        params["project_id"] = args.project_id

    client = Client.from_args(args)
    data, errors = client.get(url, params=params)
    print_response(
        data=data,
        errors=errors,
        keys=[
            "id",
            "name",
            "num_nodes",
            "crate_version",
            "project_id",
            "username",
            "fqdn",
            "channel",
        ],
        output_fmt=get_output_format(args),
    )


def clusters_deploy(args: Namespace) -> None:
    body = {
        "cluster": {
            "crate_version": args.version,
            "name": args.cluster_name,
            "password": args.password,
            "product_name": args.product_name,
            "product_tier": args.tier,
            "username": args.username,
            "channel": args.channel,
        },
        "project_id": args.project_id,
    }
    if args.unit:
        body["cluster"]["product_unit"] = args.unit
    if args.subscription_id:
        body["subscription_id"] = args.subscription_id
    _handle_edge_params(body["cluster"], args)

    client = Client.from_args(args)

    org_id = _lookup_organization_id_for_project(client, args, args.project_id)
    if not org_id:
        return

    data, errors = client.post(f"/api/v2/organizations/{org_id}/clusters/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "fqdn", "url"],
        output_fmt=get_output_format(args),
    )

    if errors or not data:
        return

    print_info("Cluster creation initiated. It may take a few minutes to complete.")

    _wait_for_completed_operation(
        client=client,
        cluster_id=data["id"],
        request_params={"type": "CREATE", "limit": 1},
    )

    # Re-fetch the cluster's info
    data, errors = client.get(f"/api/v2/clusters/{data['id']}/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "fqdn", "url"],
        output_fmt=get_output_format(args),
    )


def clusters_scale(args: Namespace) -> None:
    body = {"product_unit": args.unit}
    client = Client.from_args(args)
    data, errors = client.put(f"/api/v2/clusters/{args.cluster_id}/scale/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "num_nodes"],
        output_fmt=get_output_format(args),
    )

    if errors:
        return

    print_info(
        "Cluster scaling initiated. "
        "It may take a few minutes to complete the changes."
    )

    _wait_for_completed_operation(
        client=client,
        cluster_id=args.cluster_id,
        request_params={"type": "SCALE", "limit": 1},
    )

    # Re-fetch the cluster's info
    data, errors = client.get(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "num_nodes"],
        output_fmt=get_output_format(args),
    )


def clusters_upgrade(args: Namespace) -> None:
    body = {"crate_version": args.version}
    client = Client.from_args(args)
    data, errors = client.put(f"/api/v2/clusters/{args.cluster_id}/upgrade/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "crate_version"],
        output_fmt=get_output_format(args),
    )

    if errors:
        return

    print_info(
        "Cluster upgrade initiated. "
        "It may take a few minutes to complete the changes."
    )

    _wait_for_completed_operation(
        client=client,
        cluster_id=args.cluster_id,
        request_params={"type": "UPGRADE", "limit": 1},
    )

    # Re-fetch the cluster's info
    data, errors = client.get(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "crate_version"],
        output_fmt=get_output_format(args),
    )


@require_confirmation(
    "Are you sure you want to delete the cluster?",
    cancel_msg="Cluster deletion cancelled.",
)
def clusters_delete(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        success_message="Cluster deleted.",
        output_fmt=get_output_format(args),
    )


def clusters_restart_node(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(
        f"/api/v2/clusters/{args.cluster_id}/nodes/{args.ordinal}"
    )
    print_response(
        data=data,
        errors=errors,
        keys=["code", "status"],
        success_message=(
            "Node restarted. It may take a few minutes to complete the changes."
        ),
        output_fmt=get_output_format(args),
    )


def clusters_set_deletion_protection(args: Namespace) -> None:
    body = {"deletion_protected": args.value}
    client = Client.from_args(args)
    data, errors = client.put(
        f"/api/v2/clusters/{args.cluster_id}/deletion-protection/", body=body
    )
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "deletion_protected"],
        success_message=("Cluster deletion protection status successfully updated"),
        output_fmt=get_output_format(args),
    )


def clusters_set_suspended(args: Namespace) -> None:
    body = {"suspended": args.value}
    client = Client.from_args(args)
    data, errors = client.put(f"/api/v2/clusters/{args.cluster_id}/suspend/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "suspended"],
        output_fmt=get_output_format(args),
    )

    if errors:
        return
    print_info(
        "Updating the cluster status initiated. "
        "It may take a few minutes to complete the changes."
    )

    _wait_for_completed_operation(
        client=client,
        cluster_id=args.cluster_id,
        request_params={"type": "SUSPEND", "limit": 1},
    )

    # Re-fetch the cluster's info
    data, errors = client.get(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "suspended"],
        output_fmt=get_output_format(args),
    )


@require_confirmation("This will overwrite all existing CIDR restrictions. Continue?")
def clusters_set_ip_whitelist(args: Namespace) -> None:
    networks = args.net
    networks = networks.split(",")
    cidr = []

    for net in networks:
        if len(net) > 0:
            cidr.append({"cidr": net})

    body = {"ip_whitelist": cidr}

    client = Client.from_args(args)
    data, errors = client.put(
        f"/api/v2/clusters/{args.cluster_id}/ip-restrictions/", body=body
    )  # type: ignore
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "ip_whitelist"],
        output_fmt=get_output_format(args),
    )

    if errors:
        return
    print_info(
        "Updating the IP Network whitelist initiated. "
        "It may take a few minutes to complete the changes."
    )

    _wait_for_completed_operation(
        client=client,
        cluster_id=args.cluster_id,
        request_params={"type": "ALLOWED_CIDR_UPDATE", "limit": 1},
    )

    # Re-fetch the cluster's info
    data, errors = client.get(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "ip_whitelist"],
        output_fmt=get_output_format(args),
    )


def _disk_size_transform(field):
    disk_size = str(bitmath.Byte(field["disk_size_per_node_bytes"]).best_prefix())
    return f"Disk size: {disk_size}"


def clusters_expand_storage(args: Namespace) -> None:
    body = {
        "disk_size_per_node_bytes": args.disk_size_gb * 1024 * 1024 * 1024,
    }

    client = Client.from_args(args)
    data, errors = client.put(
        f"/api/v2/clusters/{args.cluster_id}/storage/", body=body
    )  # type: ignore
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "hardware_specs"],
        output_fmt=get_output_format(args),
        transforms={"hardware_specs": _disk_size_transform},
    )

    if errors:
        return
    print_info(
        "Cluster storage expansion initiated. "
        "It may take a few minutes to complete the changes."
    )

    _wait_for_completed_operation(
        client=client,
        cluster_id=args.cluster_id,
        request_params={"type": "EXPAND_STORAGE", "limit": 1},
    )

    # Re-fetch the cluster's info
    data, errors = client.get(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "hardware_specs"],
        output_fmt=get_output_format(args),
        transforms={"hardware_specs": _disk_size_transform},
    )


def clusters_change_product(args: Namespace) -> None:
    body = {
        "product_name": args.product_name,
    }

    client = Client.from_args(args)
    data, errors = client.put(
        f"/api/v2/clusters/{args.cluster_id}/product/", body=body
    )  # type: ignore
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "product_name"],
        output_fmt=get_output_format(args),
    )

    if errors:
        return
    print_info(
        "Changing the cluster product initiated. "
        "It may take a few minutes to complete the changes."
    )

    _wait_for_completed_operation(
        client=client,
        cluster_id=args.cluster_id,
        request_params={"type": "CHANGE_COMPUTE", "limit": 1},
    )

    # Re-fetch the cluster's info
    data, errors = client.get(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "product_name"],
        output_fmt=get_output_format(args),
    )


# We want to map the custom hardware specs to slightly nicer params in croud,
# hence this mapping here
def _handle_edge_params(body, args):
    if args.cpus:
        body.setdefault("hardware_specs", {})["cpus_per_node"] = args.cpus
    if args.disks:
        body.setdefault("hardware_specs", {})["disks_per_node"] = args.disks
    if args.disk_size_gb:
        body.setdefault("hardware_specs", {})["disk_size_per_node_bytes"] = (
            args.disk_size_gb * 1024 * 1024 * 1024
        )
    if args.disk_type:
        body.setdefault("hardware_specs", {})["disk_type"] = args.disk_type
    if args.memory_size_mb:
        body.setdefault("hardware_specs", {})["memory_per_node_bytes"] = (
            args.memory_size_mb * 1024 * 1024
        )


def _get_operation_status(client: Client, cluster_id: str, request_params: Dict):
    data, errors = client.get(
        f"/api/v2/clusters/{cluster_id}/operations/", params=request_params
    )

    if not data or len(data.get("operations", [])) == 0:
        raise AsyncOperationNotFound("Failed retrieving operation status.")

    operation = data["operations"][0]

    status = operation.get("status")
    feedback_data = operation.get("feedback_data", {})
    msg = feedback_data.get("message", None)

    return status, msg


def _wait_for_completed_operation(
    *, client: Client, cluster_id: str, request_params: Dict
):
    last_status = None
    last_msg = None
    while True:
        try:
            status, msg = _get_operation_status(
                client=client, cluster_id=cluster_id, request_params=request_params
            )
        except AsyncOperationNotFound as e:
            print_error(str(e))
            break

        if status == "SUCCEEDED":
            print_success("Operation completed.")
            break
        if status == "FAILED":
            print_error(
                "Your cluster operation has failed. "
                "Our operations team are investigating."
            )
            break

        if last_status != status or msg != last_msg:
            to_print = f"Status: {status} ({msg})" if msg else f"Status: {status}"
            print_info(to_print)
            last_status = status
            last_msg = msg

        with HALO:
            time.sleep(10)


def _lookup_organization_id_for_project(
    client: Client, args: Namespace, project_id: str
) -> Optional[str]:
    data, errors = client.get(f"/api/v2/projects/{project_id}/")
    if not data or errors:
        print_response(
            data=data,
            errors=errors,
            output_fmt=get_output_format(args),
        )
        return None
    return data.get("organization_id")



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace

from croud.api import Client
from croud.config import get_output_format
from croud.printer import print_response


def me(args: Namespace) -> None:
    """
    Prints the current logged in user
    """

    client = Client.from_args(args)
    data, errors = client.get("/api/v2/users/me/")
    print_response(
        data=data,
        errors=errors,
        keys=["email", "username"],
        output_fmt=get_output_format(args),
    )


def me_edit(args: Namespace) -> None:
    """
    Lets the user edit their data
    """

    body = {}
    if args.email:
        body["email"] = args.email

    client = Client.from_args(args)
    data, errors = client.patch("/api/v2/users/me/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["success", "status", "message"],
        output_fmt=get_output_format(args),
    )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import io
from argparse import Namespace
from unittest import mock

import pytest

from croud import __version__
from croud.parser import (
    Argument,
    CroudCliArgumentParser,
    CroudCliHelpFormatter,
    create_parser,
)
from tests.util import assert_ellipsis_match


def noop(args: Namespace):
    pass


class TestParser:
    def test_parser_instance_help(self):
        tree = {"help": "help text", "commands": {}}
        parser = create_parser(tree)

        assert isinstance(parser, CroudCliArgumentParser)
        assert parser.formatter_class == CroudCliHelpFormatter
        assert parser.prog == "croud"
        assert parser.description == "help text"
        assert parser.add_help is False

        fp = io.StringIO()
        parser.print_help(file=fp)
        fp.seek(0)
        usage = fp.read()
        assert "Show this help message and exit." in usage
        assert "Show program's version number and exit." in usage
        assert "Usage: croud" in usage
        assert "Available Commands:" in usage
        assert "Optional Arguments:" in usage
        assert "help text" in usage

    def test_parser_list_commands(self):
        tree = {
            "help": "help text",
            "commands": {
                "cmd_a": {"resolver": noop, "help": "Command A"},
                "cmd_b": {"resolver": noop, "help": "Command B"},
                "cmd_c": {"resolver": noop, "help": "Command C"},
            },
        }
        parser = create_parser(tree)

        fp = io.StringIO()
        parser.print_help(file=fp)
        fp.seek(0)
        usage = fp.read()
        assert_ellipsis_match(
            usage,
            """
Usage: croud [-h] [-v] {cmd_a,cmd_b,cmd_c}
...
Available Commands:
  cmd_a          Command A
  cmd_b          Command B
  cmd_c          Command C
...
""",
        )

    def test_parser_argument_groups(self, capsys):
        tree = {
            "help": "help text",
            "commands": {
                "cmd": {
                    "resolver": noop,
                    "extra_args": [
                        Argument("--arg-a", required=True),
                        Argument("--arg-b", required=False),
                    ],
                    "help": "Command help",
                }
            },
        }
        parser = create_parser(tree)

        with pytest.raises(SystemExit) as ex_info:
            parser.parse_args(["cmd"])
        assert ex_info.value.code == 2
        out, err = capsys.readouterr()

        assert_ellipsis_match(
            out,
            """
Usage: croud cmd [-h] --arg-a ARG_A [--arg-b ARG_B]
...
Required Arguments:
  --arg-a ARG_A
...
Optional Arguments:
...
  --arg-b ARG_B
...
""",
        )
        assert "The following arguments are required: --arg-a" in err

    def test_parser_version(self):
        tree = {"help": "help text", "commands": {}}
        parser = create_parser(tree)

        with mock.patch("sys.stdout.write") as stdout:
            with pytest.raises(SystemExit) as ex_info:
                parser.parse_args(["--version"])
            assert ex_info.value.code == 0
            assert stdout.call_args == mock.call("croud " + __version__ + "\n")

    def test_help(self):
        tree = {"help": "help text", "commands": {}}
        parser = create_parser(tree)
        with mock.patch("sys.stdout.write") as stdout:
            with pytest.raises(SystemExit) as ex_info:
                parser.parse_args(["--help"])
                stdout.assert_called_once()
                assert stdout.call_args[0][0].startswith("Usage: croud") is True
        assert ex_info.value.code == 0

    def test_no_args(self):
        tree = {"help": "help text", "commands": {}}
        parser = create_parser(tree)
        args = parser.parse_args([])
        assert args.resolver.__name__ == "print_help"

    def test_commands_without_args(self):
        tree = {"help": "help text", "commands": {"cmd": {"resolver": noop}}}
        parser = create_parser(tree)

        argv = ["cmd"]
        args = parser.parse_args(argv)
        assert args.resolver == noop
        assert args == Namespace(
            output_fmt=None, region=None, sudo=False, resolver=noop
        )

    def test_commands_with_args(self):
        tree = {
            "help": "help text",
            "commands": {
                "cmd": {
                    "resolver": noop,
                    "extra_args": [
                        Argument("-a", type=int, required=True),
                        Argument("-b", type=str, required=False),
                    ],
                }
            },
        }
        parser = create_parser(tree)

        argv = ["cmd", "-a", "1", "-b", "foo"]
        args = parser.parse_args(argv)
        assert args.resolver == noop
        assert "a" in args
        assert "b" in args

    def test_commands_with_subcommands(self):
        tree = {
            "help": "help text",
            "commands": {
                "cmd": {
                    "commands": {
                        "list": {
                            "resolver": noop,
                            "extra_args": [
                                Argument("-a", type=int, required=True),
                                Argument("-b", type=str, required=False),
                            ],
                        }
                    }
                }
            },
        }
        parser = create_parser(tree)

        argv = ["cmd", "list", "-a", "1", "-b", "foo"]
        args = parser.parse_args(argv)
        assert args.resolver == noop
        assert "a" in args
        assert "b" in args

    def test_default_arguments(self):
        tree = {"help": "help text", "commands": {"cmd": {"resolver": noop}}}
        parser = create_parser(tree)

        argv = [
            "cmd",
            "--region",
            "eastus.azure",
            "--output-fmt",
            "json",
        ]
        args = parser.parse_args(argv)
        assert args.region == "eastus.azure"
        assert args.output_fmt == "json"

    def test_default_sudo_argument(self):
        tree = {"help": "help text", "commands": {"cmd": {"resolver": noop}}}
        parser = create_parser(tree)

        argv = ["cmd", "--sudo"]
        args = parser.parse_args(argv)
        assert args.sudo is True



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import functools
import os
import platform
import subprocess
import webbrowser
from argparse import Namespace
from typing import Tuple

from croud.config import CONFIG
from croud.printer import print_error, print_info
from croud.tools.spinner import HALO


# This function was copied from the <https://github.com/Azure/azure-cli>
# project. See `LICENSE` for more information.
def can_launch_browser() -> bool:
    platform_name, release = get_platform_info()
    if is_wsl(platform_name, release) or platform_name != "linux":
        return True
    gui_env_vars = ["DESKTOP_SESSION", "XDG_CURRENT_DESKTOP", "DISPLAY"]
    result = True
    if platform_name == "linux":
        if any(os.getenv(v) for v in gui_env_vars):
            try:
                default_browser = webbrowser.get()
                if (
                    getattr(default_browser, "name", None) == "www-browser"
                ):  # text browser won't work
                    result = False
            except webbrowser.Error:
                result = False
        else:
            result = False

    return result


# This function was copied from the <https://github.com/Azure/azure-cli>
# project. See `LICENSE` for more information.
def is_wsl(platform_name: str, release: str) -> bool:
    platform_name, release = get_platform_info()
    return platform_name == "linux" and release.split("-")[-1] == "microsoft"


# This function was copied from the <https://github.com/Azure/azure-cli>
# project. See `LICENSE` for more information.
def get_platform_info() -> Tuple[str, str]:
    uname = platform.uname()
    platform_name = getattr(uname, "system", None) or uname[0]
    release = getattr(uname, "release", None) or uname[2]
    return platform_name.lower(), release.lower()


# This function was copied from the <https://github.com/Azure/azure-cli>
# project. See `LICENSE` for more information.
def open_page_in_browser(url: str) -> int:
    platform_name, release = get_platform_info()
    if is_wsl(platform_name, release):  # windows 10 linux subsystem
        try:
            return subprocess.call(
                ["cmd.exe", "/c", "start {}".format(url.replace("&", "^&"))]
            )
        except FileNotFoundError:  # WSL might be too old
            pass
    return webbrowser.open_new_tab(url)


def confirm_prompt(msg):
    msg = f"{msg.rstrip()} [yN] "
    resp = input(msg).lower()
    return resp in {"y", "yes"}


def require_confirmation(
    confirm_msg: str, cancel_msg="Command aborted!"
):  # decorator setup
    def _inner(cmd):  # decorator
        @functools.wraps(cmd)
        def _wrapper(cmd_args: Namespace):  # decorator logic
            is_confirmed = cmd_args.yes
            if not is_confirmed:
                HALO.stop()
                is_confirmed = confirm_prompt(confirm_msg)
            if is_confirmed:
                with HALO:
                    cmd(cmd_args)
            else:
                print_info(cancel_msg)

        return _wrapper

    return _inner


def org_id_config_fallback(cmd):  # decorator
    @functools.wraps(cmd)
    def _wrapper(cmd_args: Namespace):  # decorator logic
        if cmd_args.sudo and not cmd_args.org_id:
            print_error("An organization ID is required. Please pass --org-id.")
            exit(1)

        cmd_args.org_id = cmd_args.org_id or CONFIG.organization
        if not cmd_args.org_id:
            print_error("An organization ID is required. Please pass --org-id.")
            exit(1)

        cmd(cmd_args)

    return _wrapper



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import sys
from argparse import Namespace
from unittest import mock

import pytest

from croud.util import (
    can_launch_browser,
    confirm_prompt,
    get_platform_info,
    is_wsl,
    open_page_in_browser,
    org_id_config_fallback,
    require_confirmation,
)


# This function was copied from the <https://github.com/Azure/azure-cli>
# project. See `LICENSE` for more information.
@mock.patch("webbrowser.open", autospec=True)
@mock.patch("subprocess.call", autospec=True)
def test_open_page_in_browser(subprocess_call_mock, webbrowser_open_mock):
    platform_name, release = get_platform_info()
    open_page_in_browser("http://foo")
    if is_wsl(platform_name, release):
        subprocess_call_mock.assert_called_once_with(
            ["cmd.exe", "/c", "start http://foo"]
        )
    else:
        webbrowser_open_mock.assert_called_once_with("http://foo", 2)


# This function was copied from the <https://github.com/Azure/azure-cli>
# project. See `LICENSE` for more information.
@mock.patch("croud.util.get_platform_info", autospec=True)
@mock.patch("webbrowser.get", autospec=True)
def test_can_launch_browser(webbrowser_get_mock, get_platform_mock):
    # WSL is always fine
    get_platform_mock.return_value = ("linux", "4.4.0-17134-microsoft")
    result = can_launch_browser()
    assert result

    # windows is always fine
    get_platform_mock.return_value = ("windows", "10")
    result = can_launch_browser()
    assert result

    # osx is always fine
    get_platform_mock.return_value = ("darwin", "10")
    result = can_launch_browser()
    assert result

    # now tests linux
    with mock.patch("os.environ", autospec=True) as env_mock:
        # when no GUI, return false
        get_platform_mock.return_value = ("linux", "4.15.0-1014-azure")
        env_mock.get.return_value = None
        result = can_launch_browser()
        assert not result

        # when there is gui, and browser is a good one, return True
        browser_mock = mock.MagicMock()
        browser_mock.name = "goodone"
        env_mock.get.return_value = "foo"
        result = can_launch_browser()
        assert result

        # when there is gui, but the browser is text mode, return False
        browser_mock = mock.MagicMock()
        browser_mock.name = "www-browser"
        webbrowser_get_mock.return_value = browser_mock
        env_mock.get.return_value = "foo"
        result = can_launch_browser()
        assert not result


@pytest.mark.parametrize(
    "msg,response,is_confirmed",
    [
        ("test", "y", True),
        ("Other thingy", "Yes", True),
        ("But not common phrases", "Yay", False),
        ("Nope Nope Nope", "huh?", False),
        ("Srsly?", "!", False),
    ],
)
def test_confirm_prompt(msg, response, is_confirmed):
    with mock.patch("builtins.input", side_effect=[response]) as mock_input:
        assert confirm_prompt(msg) is is_confirmed
        mock_input.assert_called_once_with(f"{msg} [yN] ")


@pytest.mark.parametrize(
    "confirm_msg,cancel_msg,args_yes_value,response,output",
    [
        ("Are you sure?", None, True, "y", "Command output"),
        ("Are you sure?", None, False, "y", "Command output"),
        ("Are you sure?", "Cancelled!", False, "n", "Cancelled!"),
    ],
)
def test_require_confirmation(
    confirm_msg, cancel_msg, args_yes_value, response, output, capsys
):
    kwargs = {}
    if cancel_msg is not None:
        kwargs["cancel_msg"] = cancel_msg

    @require_confirmation(confirm_msg=confirm_msg, **kwargs)
    def command(args: Namespace):
        print("Command output")

    args = Namespace(yes=args_yes_value)
    with mock.patch("builtins.input", side_effect=[response]) as mock_input:
        command(args)
        if args_yes_value:
            mock_input.assert_not_called()
        else:
            mock_input.assert_called()

    out, err_output = capsys.readouterr()
    if err_output != "":
        assert out in err_output
    else:
        assert output in out


@pytest.mark.parametrize(
    "config_value,arg_value,expected",
    [(None, "org-1", "org-1"), ("org-1", None, "org-1"), ("org-1", "org-2", "org-2")],
)
def test_org_id_config_fallback(config_value, arg_value, expected, config, capsys):
    config.set_organization_id(config.name, config_value)

    @org_id_config_fallback
    def command(args: Namespace):
        print(args.org_id)

    args = Namespace(org_id=arg_value, sudo=False)
    command(args)
    out, _ = capsys.readouterr()
    assert expected in out


@pytest.mark.parametrize(
    "config_value,arg_value,expected",
    [(None, "org-1", "org-1"), ("org-1", None, sys.exit), ("org-1", "org-2", "org-2")],
)
def test_org_id_config_fallback_sudo(config_value, arg_value, expected, config, capsys):
    config.set_organization_id(config.name, config_value)

    @org_id_config_fallback
    def command(args: Namespace):
        print(args.org_id)

    args = Namespace(org_id=arg_value, sudo=True)

    if expected == sys.exit:
        with pytest.raises(SystemExit) as exc_info:
            command(args)
        exc_info.value.code == 1
    else:
        command(args)
        out, _ = capsys.readouterr()
        assert expected in out



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import json
import pathlib
import ssl
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread
from typing import Callable, Dict
from urllib import parse


class Response:
    __slots__ = ("bytes", "status", "headers")

    def __init__(self, *, text=None, json_data=None, status=200, headers=None):
        self.status = status
        self.headers = headers or {}
        if json_data is not None:
            self.headers.setdefault("Content-Type", "application/json")
            text = json.dumps(json_data)
        else:
            text = text or ""
            self.headers.setdefault("Content-Type", "text/html")

        self.bytes = text.encode()
        self.headers["Content-Length"] = len(self.bytes)


class FakeCrateDBCloudServer(HTTPServer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Load certificates and sign key used to simulate ssl/tls
        here = pathlib.Path(__file__)
        self.ssl_cert = here.parent / "server.crt"
        ssl_key = here.parent / "server.key"
        self.socket = ssl.wrap_socket(
            self.socket,
            keyfile=str(ssl_key),
            certfile=str(self.ssl_cert),
            server_side=True,
        )


class FakeCrateDBCloudRequestHandler(BaseHTTPRequestHandler):
    def __init__(self, *args, **kwargs):
        self.routes: Dict[str, Callable[[], Response]] = {
            "/data/data-key": self.data_data_key,
            "/data/no-key": self.data_no_key,
            "/errors/400": self.error_400,
            "/text-response": self.text_response,
            "/empty-response": self.empty_response,
            "/redirect": self.redirect,
            "/new-token": self.new_token,
            "/client-headers": self.client_headers,
        }
        super().__init__(*args, **kwargs)

    def do_GET(self):
        parsed = parse.urlparse(self.path)
        self.request_path = parsed.path
        self.query = parse.parse_qs(parsed.query)

        body_length = int(self.headers.get("Content-Length", 0))
        if body_length > 0:
            self.body = self.rfile.read(body_length)
        else:
            self.body = None

        self.cookies = dict(
            cookie.split("=", 1)
            for cookie in self.headers.get_all("cookie", [])
            if cookie
        )

        handler = self.routes.get(self.request_path, self.default_response)

        response = handler()
        self.send_response(response.status)
        for header, value in response.headers.items():
            self.send_header(header, value)
        self.end_headers()
        self.wfile.write(response.bytes)

    do_DELETE = do_GET
    do_HEAD = do_GET
    do_PATCH = do_GET
    do_POST = do_GET
    do_PUT = do_GET

    def default_response(self) -> Response:
        return Response(
            json_data={
                "body": self.body,
                "headers": dict(self.headers),  # type: ignore
                "method": self.command,
                "path": self.request_path,
                "query": self.query,
            },
            status=404,
        )

    def data_data_key(self) -> Response:
        if self.is_authorized:
            return Response(json_data={"data": {"key": "value"}}, status=200)
        return Response(status=302, headers={"Location": "/"})

    def data_no_key(self) -> Response:
        if self.is_authorized:
            return Response(json_data={"key": "value"})
        return Response(status=302, headers={"Location": "/"})

    def error_400(self) -> Response:
        if self.is_authorized:
            return Response(
                json_data={
                    "message": "Bad request.",
                    "errors": {"key": "Error on 'key'"},
                },
                status=400,
            )

        return Response(status=302, headers={"Location": "/"})

    def text_response(self) -> Response:
        if self.is_authorized:
            return Response(text="Non JSON response.", status=500)
        return Response(status=302, headers={"Location": "/"})

    def empty_response(self) -> Response:
        if self.is_authorized:
            return Response(status=204)
        return Response(status=302, headers={"Location": "/"})

    def redirect(self) -> Response:
        return Response(status=301, headers={"Location": "/?rd=%2Fredirect"})

    def new_token(self) -> Response:
        return Response(
            status=204, headers={"Set-Cookie": "session=new-token; Domain=127.0.0.1"}
        )

    def client_headers(self) -> Response:
        return Response(json_data=dict(self.headers.items()))

    @property
    def is_authorized(self) -> bool:
        if "session" in self.cookies:
            if self.cookies["session"]:
                return True
        return False

    def log_message(self, *args, **kwargs):
        # Don't log anything during tests.
        pass


class FakeCrateDBCloud:
    def __init__(self):
        self._server = FakeCrateDBCloudServer(
            ("127.0.0.1", 0), FakeCrateDBCloudRequestHandler
        )
        self._thread = Thread(target=self._server.serve_forever, daemon=True)

    def start_in_background(self) -> "FakeCrateDBCloud":
        self._thread.start()
        return self

    def wait_for_shutdown(self):
        self._server.shutdown()

    @property
    def port(self):
        return self._server.socket.getsockname()[1]

    def __enter__(self):
        return self.start_in_background()

    def __exit__(self, exc_type, exc_value, traceback):
        self.wait_for_shutdown()



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace

import bitmath

from croud.api import Client
from croud.config import get_output_format
from croud.printer import print_response


def products_list(args: Namespace) -> None:
    client = Client.from_args(args)
    url = "/api/v2/products/"
    data, errors = client.get(url, params={"kind": args.kind} if args.kind else None)
    filtered = {}
    for product in data:  # type: ignore
        # Ignore deprecated products and products that have no price associated.
        if product.get("deprecated", False):
            continue
        if not product.get("price_per_dtu_minute"):
            continue
        # group by kind/name/tier tuples to remove duplicates for providers and regions
        key = f"{product['kind']}{product['name']}{product['tier']}"
        product["vcpu_cores"] = product.get("specs", {}).get("cpu_cores")
        product["ram"] = product.get("specs", {}).get("ram_bytes", 0)
        product["min_storage"] = product.get("specs", {}).get("storage_minimum_bytes")
        product["max_storage"] = product.get("specs", {}).get("storage_maximum_bytes")
        filtered[key] = product
    print_response(
        data=list(filtered.values()),
        errors=errors,
        keys=[
            "kind",
            "name",
            "tier",
            "description",
            "scale_summary",
            "vcpu_cores",
            "ram",
            "min_storage",
            "max_storage",
        ],
        output_fmt=get_output_format(args),
        transforms={
            "ram": bytes_to_gib,
            "min_storage": bytes_to_gib,
            "max_storage": bytes_to_gib,
        },
    )


def bytes_to_gib(size_bytes):
    if not size_bytes:
        return None
    return str(bitmath.Byte(size_bytes).to_GiB())



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import sys
from argparse import Namespace

from croud.config import CONFIG, get_output_format
from croud.config.configuration import InvalidProfile
from croud.config.util import clean_dict
from croud.printer import print_error, print_format, print_info


def config_add_profile(args: Namespace) -> None:
    kwargs = {}
    if args.format:
        kwargs["format"] = args.format
    if args.region:
        kwargs["region"] = args.region
    try:
        CONFIG.add_profile(args.profile, endpoint=args.endpoint, **kwargs)
    except InvalidProfile:
        print_error(f"Failed to add profile '{args.profile}'.")
        sys.exit(1)
    else:
        print_info(f"Added profile '{args.profile}'.")


def config_current_profile(args: Namespace) -> None:
    result = {
        "name": CONFIG.name,
        "endpoint": CONFIG.endpoint,
        "format": CONFIG.format,
    }
    print_format(result, get_output_format(args))


def config_remove_profile(args: Namespace) -> None:
    try:
        CONFIG.remove_profile(args.profile)
    except InvalidProfile:
        print_error(f"Failed to remove profile '{args.profile}'.")
        print_info("Make sure the profile exists and it is not the current profile.")
        sys.exit(1)
    else:
        print_info(f"Removed profile '{args.profile}'.")


def config_set_profile(args: Namespace) -> None:
    try:
        CONFIG.use_profile(args.profile)
    except InvalidProfile:
        print_error(f"Failed to switch to profile '{args.profile}'.")
        print_info("Make sure the profile exists.")
        sys.exit(1)
    else:
        print_info(f"Switched to profile '{args.profile}'.")


def config_show(args: Namespace) -> None:
    print_info(f"Configuration file {CONFIG._file_path}")
    print_format(clean_dict(CONFIG.config), "yaml")



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import uuid
from unittest import mock

import pytest

from croud.api import Client, RequestMethod
from tests.util import assert_rest, call_command, gen_uuid

pytestmark = pytest.mark.usefixtures("config")


@mock.patch.object(Client, "request", return_value=({}, None))
def test_clusers_list(mock_request):
    call_command("croud", "clusters", "list")
    assert_rest(mock_request, RequestMethod.GET, "/api/v2/clusters/", params={})


@mock.patch.object(Client, "request", return_value=({}, None))
def test_clusers_get(mock_request):
    id = str(uuid.uuid4())
    call_command("croud", "clusters", "get", id)
    assert_rest(mock_request, RequestMethod.GET, f"/api/v2/clusters/{id}/")


@mock.patch.object(Client, "request", return_value=({}, None))
def test_clusers_list_with_project_id(mock_request):
    project_id = gen_uuid()
    call_command("croud", "clusters", "list", "--project-id", project_id)
    assert_rest(
        mock_request,
        RequestMethod.GET,
        "/api/v2/clusters/",
        params={"project_id": project_id},
    )


times_create_operations_called = 0


@mock.patch.object(Client, "request", return_value=({}, None))
def test_clusters_list_with_organization_id(mock_request):
    org_id = gen_uuid()
    call_command("croud", "clusters", "list", "--org-id", org_id)
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/organizations/{org_id}/clusters/",
        params={},
    )


@pytest.mark.parametrize("status", ["SUCCEEDED", "FAILED", None])
@mock.patch.object(Client, "request", return_value=({}, None))
@mock.patch("time.sleep")
def test_clusters_deploy(_mock_sleep, mock_request, status):
    project_id = gen_uuid()
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.POST:
            return {"id": cluster_id}, None
        if args[0] == RequestMethod.GET and "/projects/" in args[1]:
            return {"organization_id": "123"}, None
        if args[0] == RequestMethod.GET and "/operations/" in args[1]:
            if status is None:
                return None, None
            global times_create_operations_called
            if times_create_operations_called == 0:
                times_create_operations_called += 1
                return {"operations": [{"status": "SENT"}]}, None
            return {"operations": [{"status": status}]}, None
        return None, None

    mock_request.side_effect = mock_call
    call_command(
        "croud",
        "clusters",
        "deploy",
        "--product-name",
        "cratedb.az1",
        "--tier",
        "xs",
        "--unit",
        "1",
        "--project-id",
        project_id,
        "--cluster-name",
        "crate_cluster",
        "--version",
        "3.2.5",
        "--username",
        "foobar",
        "--password",
        "s3cr3t!",
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/organizations/123/clusters/",
        body={
            "cluster": {
                "crate_version": "3.2.5",
                "name": "crate_cluster",
                "password": "s3cr3t!",
                "product_name": "cratedb.az1",
                "product_tier": "xs",
                "product_unit": 1,
                "username": "foobar",
                "channel": "stable",
            },
            "project_id": project_id,
        },
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/operations/",
        params={"type": "CREATE", "limit": 1},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_clusters_deploy_fails(mock_request, capsys):
    project_id = gen_uuid()

    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.GET and "/projects/" in args[1]:
            return {"organization_id": "123"}, None
        return None, {"message": "Some Error"}

    mock_request.side_effect = mock_call
    call_command(
        "croud",
        "clusters",
        "deploy",
        "--product-name",
        "cratedb.az1",
        "--tier",
        "xs",
        "--unit",
        "1",
        "--project-id",
        project_id,
        "--cluster-name",
        "crate_cluster",
        "--version",
        "3.2.5",
        "--username",
        "foobar",
        "--password",
        "s3cr3t!",
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/organizations/123/clusters/",
        body={
            "cluster": {
                "crate_version": "3.2.5",
                "name": "crate_cluster",
                "password": "s3cr3t!",
                "product_name": "cratedb.az1",
                "product_tier": "xs",
                "product_unit": 1,
                "username": "foobar",
                "channel": "stable",
            },
            "project_id": project_id,
        },
        any_times=True,
    )
    _, err_output = capsys.readouterr()
    assert "Some Error" in err_output


@mock.patch.object(Client, "request", return_value=({}, None))
@mock.patch("time.sleep")
def test_clusters_edge(_mock_sleep, mock_request):
    project_id = gen_uuid()
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.POST:
            return {"id": cluster_id}, None
        if args[0] == RequestMethod.GET and "/projects/" in args[1]:
            return {"organization_id": "123"}, None
        return None, None

    mock_request.side_effect = mock_call
    call_command(
        "croud",
        "clusters",
        "deploy",
        "--product-name",
        "cratedb.az1",
        "--tier",
        "xs",
        "--unit",
        "1",
        "--project-id",
        project_id,
        "--cluster-name",
        "crate_cluster",
        "--version",
        "3.2.5",
        "--username",
        "foobar",
        "--password",
        "s3cr3t!",
        "--cpus",
        "1",
        "--disks",
        "1",
        "--disk-size-gb",
        "100",
        "--disk-type",
        "premium",
        "--memory-size-mb",
        "2048",
        "--subscription-id",
        "123",
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/projects/{project_id}/",
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/organizations/123/clusters/",
        body={
            "cluster": {
                "crate_version": "3.2.5",
                "name": "crate_cluster",
                "password": "s3cr3t!",
                "product_name": "cratedb.az1",
                "product_tier": "xs",
                "product_unit": 1,
                "username": "foobar",
                "channel": "stable",
                "hardware_specs": {
                    "cpus_per_node": 1.0,
                    "disks_per_node": 1,
                    "disk_size_per_node_bytes": 107_374_182_400,
                    "disk_type": "premium",
                    "memory_per_node_bytes": 2_147_483_648,
                },
            },
            "project_id": project_id,
            "subscription_id": "123",
        },
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/operations/",
        params={"type": "CREATE", "limit": 1},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )


@mock.patch.object(Client, "request", return_value=({}, None))
@mock.patch("time.sleep")
def test_clusters_deploy_no_unit(_mock_sleep, mock_request):
    project_id = gen_uuid()
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.POST:
            return {"id": cluster_id}, None
        if args[0] == RequestMethod.GET and "/projects/" in args[1]:
            return {"organization_id": "123"}, None
        return None, None

    mock_request.side_effect = mock_call
    call_command(
        "croud",
        "clusters",
        "deploy",
        "--product-name",
        "cratedb.az1",
        "--tier",
        "xs",
        "--project-id",
        project_id,
        "--cluster-name",
        "crate_cluster",
        "--version",
        "3.2.5",
        "--username",
        "foobar",
        "--password",
        "s3cr3t!",
        "--subscription-id",
        "123",
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/projects/{project_id}/",
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/organizations/123/clusters/",
        body={
            "cluster": {
                "crate_version": "3.2.5",
                "name": "crate_cluster",
                "password": "s3cr3t!",
                "product_name": "cratedb.az1",
                "product_tier": "xs",
                "username": "foobar",
                "channel": "stable",
            },
            "project_id": project_id,
            "subscription_id": "123",
        },
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/operations/",
        params={"type": "CREATE", "limit": 1},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )


@mock.patch.object(Client, "request", return_value=({}, None))
@mock.patch("time.sleep")
def test_clusters_deploy_nightly(_mock_sleep, mock_request):
    project_id = gen_uuid()
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.POST:
            return {"id": cluster_id}, None
        if args[0] == RequestMethod.GET and "/projects/" in args[1]:
            return {"organization_id": "123"}, None
        return None, None

    mock_request.side_effect = mock_call
    call_command(
        "croud",
        "clusters",
        "deploy",
        "--product-name",
        "cratedb.az1",
        "--tier",
        "xs",
        "--unit",
        "1",
        "--project-id",
        project_id,
        "--cluster-name",
        "crate_cluster",
        "--version",
        "nightly-4.1.0-20190712",
        "--username",
        "foobar",
        "--password",
        "s3cr3t!",
        "--channel",
        "nightly",
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/organizations/123/clusters/",
        body={
            "cluster": {
                "crate_version": "nightly-4.1.0-20190712",
                "name": "crate_cluster",
                "password": "s3cr3t!",
                "product_name": "cratedb.az1",
                "product_tier": "xs",
                "username": "foobar",
                "channel": "nightly",
                "product_unit": 1,
            },
            "project_id": project_id,
        },
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/operations/",
        params={"type": "CREATE", "limit": 1},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )


times_scale_operations_called = 0


@pytest.mark.parametrize("status", ["SUCCEEDED", "FAILED", None])
@mock.patch.object(Client, "request", return_value=({}, None))
@mock.patch("time.sleep")
def test_clusters_scale(_mock_sleep, mock_request: mock.Mock, status):
    unit = 1
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.GET and "/operations/" in args[1]:
            if status is None:
                return None, None
            global times_scale_operations_called

            if times_scale_operations_called == 0:
                times_scale_operations_called += 1
                return {"operations": [{"status": "SENT"}]}, None
            return {"operations": [{"status": status}]}, None
        return None, None

    mock_request.side_effect = mock_call
    call_command(
        "croud", "clusters", "scale", "--cluster-id", cluster_id, "--unit", str(unit)
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/scale/",
        body={"product_unit": unit},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/operations/",
        params={"type": "SCALE", "limit": 1},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_cluster_scale_fails(mock_request, capsys):
    unit = 1
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        return None, {"message": "Some Error"}

    mock_request.side_effect = mock_call
    call_command(
        "croud", "clusters", "scale", "--cluster-id", cluster_id, "--unit", str(unit)
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/scale/",
        body={"product_unit": unit},
    )

    _, err_output = capsys.readouterr()
    assert "Some Error" in err_output


times_upgrade_operations_called = 0


@pytest.mark.parametrize("status", ["SUCCEEDED", "FAILED", None])
@mock.patch.object(Client, "request", return_value=({}, None))
@mock.patch("time.sleep")
def test_clusters_upgrade(_mock_sleep, mock_request: mock.Mock, status):
    version = "3.2.6"
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.GET and "/operations/" in args[1]:
            if status is None:
                return None, None
            global times_upgrade_operations_called
            if times_upgrade_operations_called == 0:
                times_upgrade_operations_called += 1
                return {"operations": [{"status": "SENT"}]}, None
            return {"operations": [{"status": status}]}, None
        return None, None

    mock_request.side_effect = mock_call
    call_command(
        "croud", "clusters", "upgrade", "--cluster-id", cluster_id, "--version", version
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/upgrade/",
        body={"crate_version": version},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/operations/",
        params={"type": "UPGRADE", "limit": 1},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_cluster_upgrade_fails(mock_request, capsys):
    version = "3.2.6"
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        return None, {"message": "Some Error"}

    mock_request.side_effect = mock_call
    call_command(
        "croud", "clusters", "upgrade", "--cluster-id", cluster_id, "--version", version
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/upgrade/",
        body={"crate_version": version},
    )

    _, err_output = capsys.readouterr()
    assert "Some Error" in err_output


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_clusters_delete(mock_request, capsys):
    cluster_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["yes"]) as mock_input:
        call_command("croud", "clusters", "delete", "--cluster-id", cluster_id)
    assert_rest(mock_request, RequestMethod.DELETE, f"/api/v2/clusters/{cluster_id}/")
    mock_input.assert_called_once_with(
        "Are you sure you want to delete the cluster? [yN] "
    )

    _, err_output = capsys.readouterr()
    assert "Success" in err_output
    assert "Cluster deleted." in err_output


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_clusters_delete_flag(mock_request, capsys):
    cluster_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["y"]) as mock_input:
        call_command("croud", "clusters", "delete", "--cluster-id", cluster_id, "-y")
    assert_rest(mock_request, RequestMethod.DELETE, f"/api/v2/clusters/{cluster_id}/")
    mock_input.assert_not_called()

    _, err_output = capsys.readouterr()
    assert "Success" in err_output
    assert "Cluster deleted." in err_output


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_clusters_delete_aborted(mock_request, capsys):
    cluster_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["Nooooo"]) as mock_input:
        call_command("croud", "clusters", "delete", "--cluster-id", cluster_id)
    mock_request.assert_not_called()
    mock_input.assert_called_once_with(
        "Are you sure you want to delete the cluster? [yN] "
    )

    _, err_output = capsys.readouterr()
    assert "Cluster deletion cancelled." in err_output


@mock.patch.object(Client, "request", return_value=({}, None))
def test_clusters_restart_node(mock_request):
    ordinal = 1
    cluster_id = gen_uuid()
    call_command(
        "croud",
        "clusters",
        "restart-node",
        "--cluster-id",
        cluster_id,
        "--ordinal",
        str(ordinal),
    )
    assert_rest(
        mock_request,
        RequestMethod.DELETE,
        f"/api/v2/clusters/{cluster_id}/nodes/{ordinal}",
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_clusters_set_deletion_protection(mock_request):
    deletion_protected = True
    cluster_id = gen_uuid()
    call_command(
        "croud",
        "clusters",
        "set-deletion-protection",
        "--cluster-id",
        cluster_id,
        "--value",
        str(deletion_protected).lower(),
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/deletion-protection/",
        body={"deletion_protected": deletion_protected},
    )


times_cidr_operations_called = 0


@pytest.mark.parametrize("status", ["SUCCEEDED", "FAILED", None])
@mock.patch.object(Client, "request", return_value=({}, None))
@mock.patch("time.sleep")
def test_clusters_set_ip_whitelist(_mock_sleep, mock_request, status):
    cidr = "8.8.8.8/32,4.4.4.4/32"
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.GET and "/operations/" in args[1]:
            if status is None:
                return None, None
            global times_cidr_operations_called
            if times_cidr_operations_called == 0:
                times_cidr_operations_called += 1
                return {"operations": [{"status": "SENT"}]}, None
            return {"operations": [{"status": status}]}, None
        return None, None

    mock_request.side_effect = mock_call

    with mock.patch("builtins.input", side_effect=["yes"]) as mock_input:
        call_command(
            "croud",
            "clusters",
            "set-ip-whitelist",
            "--cluster-id",
            cluster_id,
            "--net",
            cidr,
        )
    mock_input.assert_called_once_with(
        "This will overwrite all existing CIDR restrictions. Continue? [yN] "
    )

    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/ip-restrictions/",
        body={"ip_whitelist": [{"cidr": "8.8.8.8/32"}, {"cidr": "4.4.4.4/32"}]},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/operations/",
        params={"type": "ALLOWED_CIDR_UPDATE", "limit": 1},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )


times_operations_called = 0


@pytest.mark.parametrize("status", ["SUCCEEDED", "FAILED", None])
@mock.patch.object(Client, "request", return_value=({}, None))
@mock.patch("time.sleep")
def test_clusters_reset_ip_whitelist(_mock_sleep, mock_request, status):
    cidr = ""
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.GET and "/operations/" in args[1]:
            if status is None:
                return None, None
            global times_operations_called
            if times_operations_called == 0:
                times_operations_called += 1
                return {"operations": [{"status": "SENT"}]}, None
            return {"operations": [{"status": status}]}, None
        return None, None

    mock_request.side_effect = mock_call

    call_command(
        "croud",
        "clusters",
        "set-ip-whitelist",
        "--cluster-id",
        cluster_id,
        "--net",
        cidr,
        "-y",
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/ip-restrictions/",
        body={"ip_whitelist": []},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/operations/",
        params={"type": "ALLOWED_CIDR_UPDATE", "limit": 1},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_clusters_reset_ip_whitelist_fails(mock_request, capsys):
    cidr = "8.8.8.8/32,4.4.4.4/32"
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        return None, {"message": "Some Error"}

    mock_request.side_effect = mock_call
    call_command(
        "croud",
        "clusters",
        "set-ip-whitelist",
        "--cluster-id",
        cluster_id,
        "--net",
        cidr,
        "-y",
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/ip-restrictions/",
        body={"ip_whitelist": [{"cidr": "8.8.8.8/32"}, {"cidr": "4.4.4.4/32"}]},
    )

    _, err_output = capsys.readouterr()
    assert "Some Error" in err_output


times_expand_storage_operations_called = 0


@pytest.mark.parametrize("status", ["SUCCEEDED", "FAILED", None])
@mock.patch.object(Client, "request", return_value=({}, None))
@mock.patch("time.sleep")
def test_clusters_expand_storage(_mock_sleep, mock_request: mock.Mock, status):
    disk_size_gb = 256
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.GET and "/operations/" in args[1]:
            if status is None:
                return None, None
            global times_expand_storage_operations_called

            if times_expand_storage_operations_called == 0:
                times_expand_storage_operations_called += 1
                return {"operations": [{"status": "SENT"}]}, None
            return {"operations": [{"status": status}]}, None
        return None, None

    mock_request.side_effect = mock_call
    call_command(
        "croud",
        "clusters",
        "expand-storage",
        "--cluster-id",
        cluster_id,
        "--disk-size-gb",
        str(disk_size_gb),
    )
    print("requests ", mock_request.call_args_list)
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/storage/",
        body={"disk_size_per_node_bytes": disk_size_gb * 1024 * 1024 * 1024},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/operations/",
        params={"type": "EXPAND_STORAGE", "limit": 1},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_cluster_expand_storage_fails(mock_request, capsys):
    disk_size_gb = 256
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        return None, {"message": "Some Error"}

    mock_request.side_effect = mock_call
    call_command(
        "croud",
        "clusters",
        "expand-storage",
        "--cluster-id",
        cluster_id,
        "--disk-size-gb",
        str(disk_size_gb),
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/storage/",
        body={"disk_size_per_node_bytes": disk_size_gb * 1024 * 1024 * 1024},
    )

    _, err_output = capsys.readouterr()
    assert "Some Error" in err_output


times_change_product_operations_called = 0


@pytest.mark.parametrize("status", ["SUCCEEDED", "FAILED", None])
@mock.patch.object(Client, "request", return_value=({}, None))
@mock.patch("time.sleep")
def test_clusters_change_product(_mock_sleep, mock_request: mock.Mock, status):
    product_name = "cr2"
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.GET and "/operations/" in args[1]:
            if status is None:
                return None, None
            global times_change_product_operations_called

            if times_change_product_operations_called == 0:
                times_change_product_operations_called += 1
                return {"operations": [{"status": "SENT"}]}, None
            return {"operations": [{"status": status}]}, None
        return None, None

    mock_request.side_effect = mock_call
    call_command(
        "croud",
        "clusters",
        "change-product",
        "--cluster-id",
        cluster_id,
        "--product-name",
        product_name,
    )
    print("requests ", mock_request.call_args_list)
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/product/",
        body={"product_name": product_name},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/operations/",
        params={"type": "CHANGE_COMPUTE", "limit": 1},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_clusters_change_product_fails(mock_request, capsys):
    product_name = "cr2"
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        return None, {"message": "Some Error"}

    mock_request.side_effect = mock_call
    call_command(
        "croud",
        "clusters",
        "change-product",
        "--cluster-id",
        cluster_id,
        "--product-name",
        product_name,
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/product/",
        body={"product_name": product_name},
    )

    _, err_output = capsys.readouterr()
    assert "Some Error" in err_output


times_suspend_operations_called = 0


@pytest.mark.parametrize("status", ["SUCCEEDED", "FAILED", None])
@mock.patch.object(Client, "request", return_value=({}, None))
@mock.patch("time.sleep")
def test_clusters_suspend(_mock_sleep, mock_request: mock.Mock, status):
    suspended = True
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.GET and "/operations/" in args[1]:
            if status is None:
                return None, None
            global times_suspend_operations_called

            if times_suspend_operations_called == 0:
                times_suspend_operations_called += 1
                return {"operations": [{"status": "SENT"}]}, None
            return {"operations": [{"status": status}]}, None
        return None, None

    mock_request.side_effect = mock_call
    call_command(
        "croud",
        "clusters",
        "set-suspended-state",
        "--cluster-id",
        cluster_id,
        "--value",
        str(suspended).lower(),
    )
    print("requests ", mock_request.call_args_list)
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/suspend/",
        body={"suspended": suspended},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/operations/",
        params={"type": "SUSPEND", "limit": 1},
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_cluster_suspended_fails(mock_request, capsys):
    suspended = True
    cluster_id = gen_uuid()

    def mock_call(*args, **kwargs):
        return None, {"message": "Some Error"}

    mock_request.side_effect = mock_call
    call_command(
        "croud",
        "clusters",
        "set-suspended-state",
        "--cluster-id",
        cluster_id,
        "--value",
        str(suspended).lower(),
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/suspend/",
        body={"suspended": suspended},
    )

    _, err_output = capsys.readouterr()
    assert "Some Error" in err_output
